@@ -38,11 +38,22 @@ use strict;
 use warnings;
 use re 'taint';
 
+require 5.008001;  # needs utf8::is_utf8()
+
 use Mail::SpamAssassin;
 use Mail::SpamAssassin::Constants qw(:sa);
 use Mail::SpamAssassin::HTML;
 use Mail::SpamAssassin::Logger;
 
+our($enc_utf8, $enc_w1252, $have_encode_detector);
+BEGIN {
+  eval { require Encode }
+    and do { $enc_utf8  = Encode::find_encoding('UTF-8');
+             $enc_w1252 = Encode::find_encoding('Windows-1252') };
+  eval { require Encode::Detect::Detector }
+    and do { $have_encode_detector = 1 };
+};
+
 =item new()
 
 Generates an empty Node object and returns it.  Typically only called
@@ -272,8 +283,9 @@ sub raw {
     my $fd = $self->{'raw'};
     seek($fd, 0, 0)  or die "message: cannot rewind file: $!";
 
-    my($inbuf,$nread,$raw_str); $raw_str = '';
-    while ( $nread=sysread($fd,$inbuf,16384) ) { $raw_str .= $inbuf }
+    # dbg("message: (raw) reading mime part from a temporary file");
+    my($nread,$raw_str); $raw_str = '';
+    while ( $nread=sysread($fd, $raw_str, 16384, length $raw_str) ) { }
     defined $nread  or die "error reading: $!";
     my @array = split(/^/m, $raw_str, -1);
 
@@ -308,8 +320,9 @@ sub decode {
       my $fd = $self->{'raw'};
       seek($fd, 0, 0)  or die "message: cannot rewind file: $!";
 
-      my($inbuf,$nread,$raw_str); $raw = '';
-      while ( $nread=sysread($fd,$inbuf,16384) ) { $raw .= $inbuf }
+      # dbg("message: (decode) reading mime part from a temporary file");
+      my($nread,$raw_str); $raw = '';
+      while ( $nread=sysread($fd, $raw, 16384, length $raw) ) { }
       defined $nread  or die "error reading: $!";
 
       dbg("message: empty message read from a temp file")  if $raw eq '';
@@ -383,29 +396,189 @@ sub _html_render {
   return 0;
 }
 
+# Decode character set of a given text to perl characters (Unicode),
+# then encode into UTF-8 octets if requested.
+#
 sub _normalize {
-  my ($self, $data, $charset) = @_;
-  return $data unless $self->{normalize};
+  my $self = $_[0];
+# my $data = $_[1];  # avoid copying large strings
+  my $charset_declared = $_[2];
+  my $return_decoded = $_[3];  # true: Unicode characters, false: UTF-8 octets
 
-  my $detected = Encode::Detect::Detector::detect($data);
+  return $_[1]  unless $self->{normalize} && $enc_utf8;
 
-  my $converter;
+  warn "message: _normalize() was given characters, expected bytes: $_[1]\n"
+    if utf8::is_utf8($_[1]);
 
-  if ($charset && $charset !~ /^us-ascii$/i &&
-      ($detected || 'none') !~ /^(?:UTF|EUC|ISO-2022|Shift_JIS|Big5|GB)/i) {
-      dbg("message: Using labeled charset $charset");
-      $converter = Encode::find_encoding($charset);
+  # workaround for Encode::decode taint laundering bug [rt.cpan.org #84879]
+  my $data_taint = substr($_[1], 0, 0);  # empty string, tainted like $data
+
+  if (!defined $charset_declared || $charset_declared eq '') {
+    $charset_declared = 'us-ascii';
   }
 
-  $converter = Encode::find_encoding($detected) unless $converter || !defined($detected);
+  # number of characters with code above 127
+  my $cnt_8bits = $_[1] =~ tr/\x00-\x7F//c;
 
-  return $data unless $converter;
+  if (!$cnt_8bits &&
+      $charset_declared =~
+        /^(?: (?:US-)?ASCII | ANSI[_ ]? X3\.4- (?:1986|1968) |
+              ISO646-US )\z/xsi)
+  { # declared as US-ASCII (a.k.a. ANSI X3.4-1986) and it really is
+    dbg("message: kept, charset is US-ASCII as declared");
+    return $_[1];  # is all-ASCII, no need for decoding
+  }
 
-  dbg("message: Converting...");
+  if (!$cnt_8bits &&
+      $charset_declared =~
+        /^(?: ISO[ -]?8859 (?: - \d{1,2} )? | Windows-\d{4} |
+              UTF-?8 | (KOI8|EUC)-[A-Z]{1,2} |
+              Big5 | GBK | GB[ -]?18030 (?:-20\d\d)? )\z/xsi)
+  { # declared as extended ASCII, but it is actually a plain 7-bit US-ASCII
+    dbg("message: kept, charset is US-ASCII, declared %s", $charset_declared);
+    return $_[1];  # is all-ASCII, no need for decoding
+  }
 
-  my $rv = $converter->decode($data, 0);
-  utf8::downgrade($rv, 1);
-  return $rv
+  # Try first to strictly decode based on a declared character set.
+
+  my $rv;
+  if ($charset_declared =~ /^UTF-?8\z/i) {
+    # attempt decoding as strict UTF-8  (flags: FB_CROAK | LEAVE_SRC)
+    if (eval { $rv = $enc_utf8->decode($_[1], 1|8); defined $rv }) {
+      dbg("message: decoded as declared charset UTF-8");
+      return $_[1]  if !$return_decoded;
+      $rv .= $data_taint;  # carry taintedness over, avoid Encode bug
+      return $rv;  # decoded
+    } else {
+      dbg("message: failed decoding as declared charset UTF-8");
+    };
+
+  } elsif ($cnt_8bits &&
+           eval { $rv = $enc_utf8->decode($_[1], 1|8); defined $rv }) {
+    dbg("message: decoded as charset UTF-8, declared %s", $charset_declared);
+    return $_[1]  if !$return_decoded;
+    $rv .= $data_taint;  # carry taintedness over, avoid Encode bug
+    return $rv;  # decoded
+
+  } elsif ($charset_declared =~ /^(?:US-)?ASCII\z/i) {
+    # declared as US-ASCII but contains 8-bit characters, makes no sense
+    # to attempt decoding first as strict US-ASCII as we know it would fail
+
+  } else {
+    # try decoding as a declared character set
+
+    # ->  http://en.wikipedia.org/wiki/Windows-1252
+    # Windows-1252 character encoding is a superset of ISO 8859-1, but differs
+    # from the IANA's ISO-8859-1 by using displayable characters rather than
+    # control characters in the 80 to 9F (hex) range. [...]
+    # It is very common to mislabel Windows-1252 text with the charset label
+    # ISO-8859-1. A common result was that all the quotes and apostrophes
+    # (produced by "smart quotes" in word-processing software) were replaced
+    # with question marks or boxes on non-Windows operating systems, making
+    # text difficult to read. Most modern web browsers and e-mail clients
+    # treat the MIME charset ISO-8859-1 as Windows-1252 to accommodate
+    # such mislabeling. This is now standard behavior in the draft HTML 5
+    # specification, which requires that documents advertised as ISO-8859-1
+    # actually be parsed with the Windows-1252 encoding.
+    #
+    my($chset, $decoder);
+    if ($charset_declared =~ /^(?: ISO-?8859-1 | Windows-1252 | CP1252 )\z/xi) {
+      $chset = 'Windows-1252'; $decoder = $enc_w1252;
+    } else {
+      $chset = $charset_declared; $decoder = Encode::find_encoding($chset);
+      if (!$decoder && $chset =~ /^GB[ -]?18030(?:-20\d\d)?\z/i) {
+        $decoder = Encode::find_encoding('GBK');  # a subset of GB18030
+        dbg("message: no decoder for a declared charset %s, using GBK",
+            $chset)  if $decoder;
+      }
+    }
+    if (!$decoder) {
+      dbg("message: failed decoding, no decoder for a declared charset %s",
+          $chset);
+    } else {
+      eval { $rv = $decoder->decode($_[1], 1|8) };  # FB_CROAK | LEAVE_SRC
+      if (lc $chset eq lc $charset_declared) {
+        dbg("message: %s as declared charset %s",
+            defined $rv ? 'decoded' : 'failed decoding', $charset_declared);
+      } else {
+        dbg("message: %s as charset %s, declared %s",
+            defined $rv ? 'decoded' : 'failed decoding',
+            $chset, $charset_declared);
+      }
+    }
+  }
+
+  # If the above failed, check if it is US-ASCII, possibly extended by few
+  # NBSP or SHY characters from ISO-8859-* or Windows-1252, or containing
+  # some popular punctuation or special characters from Windows-1252 in
+  # the \x80-\x9F range (which is unassigned in ISO-8859-*).
+  # Note that Windows-1252 is a proper superset of ISO-8859-1.
+  #
+  if (!defined $rv && !$cnt_8bits) {
+    dbg("message: kept, guessed charset is US-ASCII, declared %s",
+        $charset_declared);
+    return $_[1];  # is all-ASCII, no need for decoding
+
+  } elsif (!defined $rv && $enc_w1252 &&
+      #             ASCII  NBSP (c) SHY  '   "  ...   '".-   TM
+      $_[1] !~ tr/\x00-\x7F\xA0\xA9\xAD\x82\x84\x85\x91-\x97\x99//c)
+  { # ASCII + NBSP + SHY + some punctuation characters
+    # NBSP (A0) and SHY (AD) are at the same position in ISO-8859-* too
+    # consider also: AE (r), 80 Euro
+    eval { $rv = $enc_w1252->decode($_[1], 1|8) };  # FB_CROAK | LEAVE_SRC
+    # the above can't fail, but keep code general just in case
+    dbg("message: %s as guessed charset %s, declared %s",
+        defined $rv ? 'decoded' : 'failed decoding',
+        'Windows-1252', $charset_declared);
+  }
+
+  # If we were unsuccessful so far, try some guesswork
+  # based on Encode::Detect::Detector .
+
+  if (defined $rv) {
+    # done, no need for guesswork
+  } elsif (!$have_encode_detector) {
+    dbg("message: Encode::Detect::Detector not available, declared %s failed",
+        $charset_declared);
+  } else {
+    my $charset_detected = Encode::Detect::Detector::detect($_[1]);
+    if ($charset_detected && lc $charset_detected ne lc $charset_declared) {
+      my $decoder = Encode::find_encoding($charset_detected);
+      if (!$decoder && $charset_detected =~ /^GB[ -]?18030(?:-20\d\d)?\z/i) {
+        $decoder = Encode::find_encoding('GBK');  # a subset of GB18030
+        dbg("message: no decoder for a detected charset %s, using GBK",
+            $charset_detected)  if $decoder;
+      }
+      if (!$decoder) {
+        dbg("message: failed decoding, no decoder for a detected charset %s",
+            $charset_detected);
+      } else {
+        eval { $rv = $decoder->decode($_[1], 1|8) };  # FB_CROAK | LEAVE_SRC
+        dbg("message: %s as detected charset %s, declared %s",
+            defined $rv ? 'decoded' : 'failed decoding',
+            $charset_detected, $charset_declared);
+      }
+    }
+  }
+
+  if (!defined $rv) {  # all decoding attempts failed so far, probably garbage
+    # go for Windows-1252 which can't fail
+    eval { $rv = $enc_w1252->decode($_[1]) };
+    dbg("message: %s as last-resort charset %s, declared %s",
+        defined $rv ? 'decoded' : 'failed decoding',
+        'Windows-1252', $charset_declared);
+  }
+
+  if (!defined $rv) {  # just in case - all decoding attempts failed so far
+    return $_[1];  # garbage-in / garbage-out, return unchanged octets
+  }
+  # decoding octets to characters was successful
+  if (!$return_decoded) {
+    # utf8::encode() is much faster than $enc_utf8->encode on utf8-flagged arg
+    utf8::encode($rv);  # encode Unicode characters to UTF-8 octets
+  }
+  $rv .= $data_taint;  # carry taintedness over, avoid Encode bug
+  return $rv;
 }
 
 =item rendered()
@@ -428,19 +601,42 @@ sub rendered {
     # text/x-aol is ignored here, but looks like text/html ...
     return(undef,undef) unless ( $self->{'type'} =~ /^text\/(?:plain|html)$/i );
 
-    my $text = $self->_normalize($self->decode(), $self->{charset});
-    my $raw = length($text);
+    my $text = $self->decode;  # QP and Base64 decoding, bytes
+    my $text_len = length($text);  # num of bytes in original charset encoding
 
     # render text/html always, or any other text|text/plain part as text/html
     # based on a heuristic which simulates a certain common mail client
-    if ($raw > 0 && ($self->{'type'} =~ m@^text/html$@i ||
-		     ($self->{'type'} =~ m@^text/plain$@i &&
-		      _html_render(substr($text, 0, 23)))))
+    if ($text ne '' && ($self->{'type'} =~ m{^text/html$}i ||
+		        ($self->{'type'} =~ m{^text/plain$}i &&
+		         _html_render(substr($text, 0, 23)))))
     {
       $self->{rendered_type} = 'text/html';
 
-      my $html = Mail::SpamAssassin::HTML->new();	# object
-      $html->parse($text);				# parse+render text
+      # will input text to HTML::Parser be provided as Unicode characters?
+      my $character_semantics = 0;  # $text is in bytes
+      if ($self->{normalize} && $enc_utf8) {  # charset decoding requested
+        # Provide input to HTML::Parser as Unicode characters
+        # which avoids a HTML::Parser bug in utf8_mode
+        #   https://rt.cpan.org/Public/Bug/Display.html?id=99755
+        # Avoid unnecessary step of encoding-then-decoding by telling
+        # subroutine _normalize() to return Unicode text.  See Bug 7133
+        #
+        $character_semantics = 1;  # $text will be in characters
+        $text = $self->_normalize($text, $self->{charset}, 1); # bytes to chars
+      } elsif (!defined $self->{charset} ||
+               $self->{charset} =~ /^(?:US-ASCII|UTF-8)\z/i) {
+        # With some luck input can be interpreted as UTF-8, do not warn.
+        # It is still possible to hit the HTML::Parses utf8_mode bug however.
+      } else {
+        dbg("message: 'normalize_charset' is off, encoding will likely ".
+            "be misinterpreted; declared charset: %s", $self->{charset});
+      }
+      # the 0 requires decoded HTML results to be in bytes (not characters)
+      my $html = Mail::SpamAssassin::HTML->new($character_semantics,0); # object
+
+      $html->parse($text);  # parse+render text
+
+      # resulting HTML-decoded text is in bytes, likely encoded as UTF-8
       $self->{rendered} = $html->get_rendered_text();
       $self->{visible_rendered} = $html->get_rendered_text(invisible => 0);
       $self->{invisible_rendered} = $html->get_rendered_text(invisible => 1);
@@ -449,15 +645,20 @@ sub rendered {
       # end-of-document result values that require looking at the text
       my $r = $self->{html_results};	# temporary reference for brevity
 
-      # count the number of spaces in the rendered text
-      my $rt = pack "C0A*", $self->{rendered};
-      my $space = ($rt =~ tr/ \t\n\r\x0b\xa0/ \t\n\r\x0b\xa0/);
-      $r->{html_length} = length($rt);
+      # count the number of spaces in the rendered text (likely UTF-8 octets)
+      my $space = $self->{rendered} =~ tr/ \t\n\r\x0b//;
+      # we may want to add the count of other Unicode whitespace characters
 
+      $r->{html_length} = length $self->{rendered};  # bytes (likely UTF-8)
       $r->{non_space_len} = $r->{html_length} - $space;
-      $r->{ratio} = ($raw - $r->{html_length}) / $raw;
+      $r->{ratio} = ($text_len - $r->{html_length}) / $text_len  if $text_len;
     }
-    else {
+
+    else {  # plain text
+      if ($self->{normalize} && $enc_utf8) {
+        # request transcoded result as UTF-8 octets!
+        $text = $self->_normalize($text, $self->{charset}, 0);
+      }
       $self->{rendered_type} = $self->{type};
       $self->{rendered} = $self->{'visible_rendered'} = $text;
       $self->{'invisible_rendered'} = '';
@@ -579,7 +780,7 @@ sub __decode_header {
     # not possible since the input has already been limited to 'B' and 'Q'
     die "message: unknown encoding type '$cte' in RFC2047 header";
   }
-  return $self->_normalize($data, $encoding);
+  return $self->_normalize($data, $encoding, 0);  # transcode to UTF-8 octets
 }
 
 # Decode base64 and quoted-printable in headers according to RFC2047.
@@ -593,30 +794,32 @@ sub _decode_header {
   $header_field_body =~ s/\n[ \t]+/\n /g;
   $header_field_body =~ s/\015?\012//gs;
 
-  # Bug 6945: some header fields must not be processed for MIME encoding
   if ($header_field_name =~
        /^ (?: (?: Received | (?:Resent-)? (?: Message-ID | Date ) |
                   MIME-Version | References | In-Reply-To ) \z
             | (?: List- | Content- ) ) /xsi ) {
-    return $header_field_body;
+    # Bug 6945: some header fields must not be processed for MIME encoding
+
+  } else {
+    local($1,$2,$3);
+
+    # Multiple encoded sections must ignore the interim whitespace.
+    # To avoid possible FPs with (\s+(?==\?))?, look for the whole RE
+    # separated by whitespace.
+    1 while $header_field_body =~
+              s{ ( = \? [A-Za-z0-9_-]+ \? [bqBQ] \? [^?]* \? = ) \s+
+                 ( = \? [A-Za-z0-9_-]+ \? [bqBQ] \? [^?]* \? = ) }
+               {$1$2}xsg;
+
+    # transcode properly encoded RFC 2047 substrings into UTF-8 octets,
+    # leave everything else unchanged as it is supposed to be UTF-8 (RFC 6532)
+    # or plain US-ASCII
+    $header_field_body =~
+      s{ (?: = \? ([A-Za-z0-9_-]+) \? ([bqBQ]) \? ([^?]*) \? = ) }
+       { $self->__decode_header($1, uc($2), $3) }xsge;
   }
 
-  # multiple encoded sections must ignore the interim whitespace.
-  # to avoid possible FPs with (\s+(?==\?))?, look for the whole RE
-  # separated by whitespace.
-  1 while ($header_field_body =~ s/(=\?[\w_-]+\?[bqBQ]\?[^?]+\?=)\s+(=\?[\w_-]+\?[bqBQ]\?[^?]+\?=)/$1$2/g);
-
-# unless ($header_field_body =~
-#	  s/=\?([\w_-]+)\?([bqBQ])\?([^?]+)\?=/$self->__decode_header($1, uc($2), $3)/ge) {
-#   $header_field_body = $self->_normalize($header_field_body);
-# }
-
-  # Bug 6945: header fields have no inherent character set;
-  # only decode and normalize properly encoded RFC 2047 substrings,
-  # leave the rest as plain octets
-  $header_field_body =~ s{ = \? ([A-Za-z0-9_-]+) \? ([bqBQ]) \? ([^?]+) \? = }
-                         { $self->__decode_header($1, uc($2), $3) }xge;
-
+# dbg("message: _decode_header %s: %s", $header_field_name, $header_field_body);
   return $header_field_body;
 }
 
