@@ -49,6 +49,7 @@ package Mail::SpamAssassin::PersistentAddrList;
 use strict;
 use warnings;
 use bytes;
+use re 'taint';
 
 use vars qw{
   @ISA
@@ -110,7 +111,7 @@ a C<count> key and a C<totscore> key.
 =cut 
 
 sub get_addr_entry {
-  my ($self, $addr) = @_;
+  my ($self, $addr, $signedby) = @_;
   my $entry = { };
   die "auto-whitelist: unimplemented base method";	# override this
   return $entry;
