@@ -68,6 +68,7 @@ sub new {
   $self->register_eval_rule("check_ratware_envelope_from");
   $self->register_eval_rule("gated_through_received_hdr_remover");
   $self->register_eval_rule("received_within_months");
+  $self->register_eval_rule("check_equal_from_domains");
 
   return $self;
 }
@@ -79,10 +80,6 @@ sub compile_now_start {
   $self->word_is_in_dictionary("aba");
 }
 
-# sad but true. sort it out, sysadmins!
-my $CCTLDS_WITH_LOTS_OF_OPEN_RELAYS = qr{(?:kr|cn|cl|ar|hk|il|th|tw|sg|za|tr|ma|ua|in|pe|br)};
-my $ROUND_THE_WORLD_RELAYERS = qr{(?:net|com|ca)};
-
 sub check_for_fake_aol_relay_in_rcvd {
   my ($self, $pms) = @_;
   local ($_);
@@ -142,7 +139,7 @@ sub check_for_unique_subject_id {
   my ($self, $pms) = @_;
   local ($_);
   $_ = lc $pms->get('Subject');
-  study;
+  study;  # study is a no-op since perl 5.16.0, eliminating related bugs
 
   my $id = 0;
   if (/[-_\.\s]{7,}([-a-z0-9]{4,})$/
@@ -270,26 +267,27 @@ sub check_illegal_chars {
 
   $header .= ":raw" unless ($header eq "ALL" || $header =~ /:raw$/);
   my $str = $pms->get($header);
-  return 0 unless $str ne '';
+  return 0 if !defined $str || $str eq '';
 
   # avoid overlap between tests
   if ($header eq "ALL") {
     # fix continuation lines, then remove Subject and From
     $str =~ s/\n[ \t]+/  /gs;
-    $str =~ s/^(?:Subject|From):.*$//gm;
+    $str =~ s/^(?:Subject|From):.*$//gmi;
   }
 
   # count illegal substrings (RFC 2045)
-  my $illegal = () = ($str =~ /[\x00-\x08\x0b\x0c\x0e-\x1f\x7f-\xff]/g);
+  # (non-ASCII + C0 controls except TAB, NL, CR)
+  my $illegal = $str =~ tr/\x00-\x08\x0b\x0c\x0e-\x1f\x7f-\xff//;
 
   # minor exemptions for Subject
-  if ($header eq "Subject:raw") {
+  if ($illegal > 0 && lc $header eq "subject:raw") {
     # only exempt a single cent sign, pound sign, or registered sign
-    my $exempt = () = ($str =~ /[\xa2\xa3\xae]/g);
+    my $exempt = $str =~ tr/\xa2\xa3\xae//;
     $illegal-- if $exempt == 1;
   }
 
-  return 0 if (length($str) == 0);
+  return 0 if $str eq '';
   return (($illegal / length($str)) >= $ratio && $illegal >= $count);
 }
 
@@ -769,11 +767,9 @@ sub _get_received_header_times {
     unshift @received, (shift @local);
   }
 
-  my $rcvd;
-
   if (scalar(@local)) {
     my (@fetchmail_times);
-    foreach $rcvd (@local) {
+    foreach my $rcvd (@local) {
       if ($rcvd =~ m/(\s.?\d+ \S\S\S \d+ \d+:\d+:\d+ \S+)/) {
 	my $date = $1;
         dbg2("eval: trying Received fetchmail header date for real time: $date");
@@ -793,7 +789,7 @@ sub _get_received_header_times {
   }
 
   my (@header_times);
-  foreach $rcvd (@received) {
+  foreach my $rcvd (@received) {
     if ($rcvd =~ m/(\s.?\d+ \S\S\S \d+ \d+:\d+:\d+ \S+)/) {
       my $date = $1;
       dbg2("eval: trying Received header date for real time: $date");
@@ -898,6 +894,7 @@ sub subject_is_all_caps {
 
    $subject =~ s/^\s+//;
    $subject =~ s/\s+$//;
+   $subject =~ s/^(?:(?:Re|Fwd|Fw|Aw|Antwort|Sv):\s*)+//i;  # Bug 6805
    return 0 if $subject !~ /\s/;	# don't match one word subjects
    return 0 if (length $subject < 10);  # don't match short subjects
    $subject =~ s/[^a-zA-Z]//g;		# only look at letters
@@ -976,7 +973,7 @@ sub check_messageid_not_usable {
 
   # Lyris eats message-ids.  also some ezmlm, I think :(
   $_ = $pms->get("List-Unsubscribe");
-  return 1 if (/<mailto:(?:leave-\S+|\S+-unsubscribe)\@\S+>$/);
+  return 1 if (/<mailto:(?:leave-\S+|\S+-unsubscribe)\@\S+>$/i);
 
   # ezmlm again
   if($self->gated_through_received_hdr_remover($pms)) { return 1; }
@@ -1040,9 +1037,9 @@ sub check_ratware_envelope_from {
 
   if ($to =~ /^([^@]+)@(.+)$/) {
     my($user,$dom) = ($1,$2);
-    $dom = Mail::SpamAssassin::Util::RegistrarBoundaries::trim_domain($dom);
+    $dom = $self->{main}->{registryboundaries}->trim_domain($dom);
     return unless
-        (Mail::SpamAssassin::Util::RegistrarBoundaries::is_domain_valid($dom));
+        ($self->{main}->{registryboundaries}->is_domain_valid($dom));
 
     return 1 if ($from =~ /\b\Q$dom\E.\Q$user\E@/i);
   }
@@ -1050,6 +1047,33 @@ sub check_ratware_envelope_from {
   return 0;
 }
 
+# ADDED FROM BUG 6487
+sub check_equal_from_domains {
+  my ($self, $pms) = @_;
+
+  my $from = $pms->get('From:addr');
+  my $envfrom = $pms->get('EnvelopeFrom:addr');
+
+  local $1;
+  my $fromdomain = '';
+  #Revised regexp from 6487 comment 3
+  $fromdomain = $1  if $from =~ /\@([^@]*)\z/;
+  $fromdomain =~ s/^.+\.([^\.]+\.[^\.]+)$/$1/;
+  return 0 if $fromdomain eq '';
+
+  my $envfromdomain = '';
+  $envfromdomain = $1  if $envfrom =~ /\@([^@]*)\z/;
+  $envfromdomain =~ s/^.+\.([^\.]+\.[^\.]+)$/$1/;
+  return 0 if $envfromdomain eq '';
+
+  dbg("eval: From 2nd level domain: $fromdomain, EnvelopeFrom 2nd level domain: $envfromdomain");
+  
+  return 1 if lc($fromdomain) ne lc($envfromdomain);
+
+  return 0;
+}
+
+
 ###########################################################################
 
 # support eval-test verbose debugs using "-Deval"
